#  $Id: config_tmpl.tk,v 4.9 2002/04/09 11:48:48 katie Exp $
# Copyright (c) 2001,2002                        RIPE NCC
#
# All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the author not be
# used in advertising or publicity pertaining to distribution of the
# software without specific, written prior permission.
#
# THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
# ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
# AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
# AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#
#  Copyright (c) 1994 by the University of Southern California
#  All rights reserved.
#
#  Permission to use, copy, modify, and distribute this software and its
#  documentation in source and binary forms for lawful non-commercial
#  purposes and without fee is hereby granted, provided that the above
#  copyright notice appear in all copies and that both the copyright
#  notice and this permission notice appear in supporting documentation,
#  and that any documentation, advertising materials, and other materials
#  related to such distribution and use acknowledge that the software was
#  developed by the University of Southern California, Information
#  Sciences Institute. The name of the USC may not be used to endorse or
#  promote products derived from this software without specific prior
#  written permission.
#
#  THE UNIVERSITY OF SOUTHERN CALIFORNIA DOES NOT MAKE ANY
#  REPRESENTATIONS ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY
#  PURPOSE.  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
#  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE,
#  TITLE, AND NON-INFRINGEMENT.
#
#  IN NO EVENT SHALL USC, OR ANY OTHER CONTRIBUTOR BE LIABLE FOR ANY
#  SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES, WHETHER IN CONTRACT, TORT,
#  OR OTHER FORM OF ACTION, ARISING OUT OF OR IN CONNECTION WITH, THE USE
#  OR PERFORMANCE OF THIS SOFTWARE.
#
#  Questions concerning this software should be directed to
#  ratoolset@isi.edu.
#
#  Author(s): Cengiz Alaettinoglu <cengiz@ISI.EDU>

toplevel .config_tmpl
wm title .config_tmpl {ROE Template Configuration}

label .config_tmpl.add_label -text {Add template:} -bd 2
pack  .config_tmpl.add_label -side top -anchor w
text  .config_tmpl.add_text  -relief sunken -bd 2 -width 40 -height 8
pack  .config_tmpl.add_text  -side top -fill both -expand true -anchor nw 

label .config_tmpl.upd_label -text {Update template:} -bd 2
pack  .config_tmpl.upd_label -side top -anchor w
text  .config_tmpl.upd_text  -relief sunken -bd 2 -width 40 -height 2
pack  .config_tmpl.upd_text  -side top -fill both -expand true -anchor nw 

label .config_tmpl.del_label -text {Delete template:} -bd 2
pack  .config_tmpl.del_label -side top -anchor w
text  .config_tmpl.del_text  -relief sunken -bd 2 -width 40 -height 2
pack  .config_tmpl.del_text  -side top -fill both -expand true -anchor nw 

button .config_tmpl.apply -border 1 -text Done \
	    -command {roe_config_tmpl_apply; destroy .config_tmpl}
pack .config_tmpl.apply -side left -padx 2 -pady 2 -ipadx 2

button .config_tmpl.cancel -text Cancel -command {destroy .config_tmpl}
pack .config_tmpl.cancel -side left -pady 2 -ipadx 2

focus .config_tmpl
